/*
	$Id: DalDatabase.h 1193 2018-11-20 11:58:19Z john $

	This file is part of the RFC DAL.

	The RFC DAL is Copyright (c) 2004 - 2007, The 4c Company.

	The RFC DAL is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, version 2 of the License.

	The RFC DAL is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with This code.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __DalDatabase_h__
#define __DalDatabase_h__

#include <librfcdal/SqlGlobal.h>
#include <librfcdal/SqlDatabase.h>
#include <librfcdal/SqlSchema.h>
#include <librfcdal/SqlValue.h>
#include <librfcdal/SqlLanguage.h>
#include <librfcdal/SqlColumnInfo.h>

#include <librfcdal/DalDatabaseException.h>

#include <vector>
#include <map>
#include <string>

namespace RfcDal
{
	class SqlModifyStatement;
	class SqlQueryStatement;
	class DalDynamic;

	class RFCDAL_EXPORT DalDatabase
	{
		public:
			class Row : public std::map<std::string, SqlValue>
			{
			};

			class Result : public std::vector<Row>
			{
			};

			class RFCDAL_EXPORT ConnectionParameter
			{
				public:
					// Requirement:
					// Optional : the parameter is optional
					// Mandatory : the parameter is mandatory
					// reqDatabase : the parameter is a database identifier
					// reqFilename : the parameter is a filename, with or without path
					// reqHost : the parameter is a host name or IP address
					// reqPassword : the parameter is a user password
					// reqSchema : the parameter is a database schema _not_ associated with any specified user
					// reqSchemaUser : the parameter is a database schmema _and_ user
					// reqUser : the parameter is a user ID _not_ associated with a schema
					// reqOther : the parameter is of un-specified type
					enum Requirement { reqDatabaseMandatory = 'D',
										reqFilenameMandatory = 'F',
										reqHostMandatory = 'H',
										reqPasswordMandatory = 'P',
										reqSchemaMandatory = 'S',
										reqSchemaUserMandatory = 'T',
										reqUserMandatory = 'U',
										reqOtherMandatory = 'Z',
										reqDatabaseOptional = 'd',
										reqFilenameOptional = 'f',
										reqHostOptional = 'h',
										reqPasswordOptional = 'p',
										reqSchemaOptional = 's',
										reqSchemaUserOptional = 't',
										reqUserOptional = 'u',
										reqOtherOptional = 'z' };

					static const int noGroup;

				private:
					std::string _strParameter;
					Requirement _requirement;
					SqlLanguage::CaseSensitivity _caseSensitivity;
					// _nExclusiveGroup == noGroup (-1), the item is not in an exclusive group
					int _nExclusiveGroup;

				public:
					ConnectionParameter( const std::string& strParameter = std::string(), Requirement requirement = reqOtherOptional,
											SqlLanguage::CaseSensitivity caseSensitivity = SqlLanguage::caseInsensitive,
											int nExclusiveGroup = noGroup );
					ConnectionParameter( const ConnectionParameter& param );
					virtual ~ConnectionParameter();

					const ConnectionParameter& operator=( const ConnectionParameter& param );

					const std::string& parameter() const;
					Requirement requirement() const;
					SqlLanguage::CaseSensitivity caseSensitivity() const;
					int exclusiveGroup() const;

					bool isMandatory() const;
					bool isUserId() const;
					bool isPassword() const;
					bool isDatabase() const;
					bool isSchema() const;
					bool isHost() const;
					bool isFilename() const;
			};

			class ConnectionParameterList : public std::vector<ConnectionParameter>
			{
			};

			class ConnectionParameterMap : public std::map<std::string, std::string>
			{
			};

		private:
			bool _bIsConnected;
			SqlLanguage _language;

		protected:
			static ConnectionParameterMap _mapDefault;

		public:
			DalDatabase();
			virtual ~DalDatabase();

			/*
				connectionParameters
				returns a vector of named ConnectionParameter objects that identify the named parameters to be used in the connectX
				functions, along with whether they are optional or required.
			*/
			virtual const ConnectionParameterList& connectionParameters() const = 0;

			/*
				connectX
				Connect strings are of the form <parameter name>=<value><delimiter>[...], for example:
				dsn=Test1;userid=testuser;password=mypass
				This function will call connectStringToMap before calling the second overload of the connectX function,
				passing in the ConnectionParameterMap returned by connectStringToMap. If the optional map argument is
				used, upon return, it will contain the parameters parsed out of the connect string.
				If the connect string contains a semi-colon (;), it will assume that the delimiter is semi-colon,
				otherwise it will assume that the delimiter is colon (:). If the parameters can contain either of these
				values, the options are either to build the ConnectionParameterMap manually or to use the function
				connectStringToMap, specifying the delimiter.
				For a list of the acceptable parameters for any given database adaptor, see the connectionParameters
				function.
			*/
			virtual bool connectX( const std::string& strConnect, ConnectionParameterMap& map = _mapDefault );

			/*
				connectX
				This function is implemented by the specific database adaptors and takes as input the
				ConnectionParameterMap returned by connectStringToMap.
			*/
			virtual bool connectX( ConnectionParameterMap& map ) = 0;

			/*
				isConnected
				returns true if connectX or connect has succeeded and false if they have not. This does not mean
				that the connection is active as many data sources have inactivity timeouts.
			*/ 
			bool isConnected() const;

			/*
				disconnect
				Disconnect the adaptor from the data source.
			*/
			virtual void disconnect();

			/*
				language
				returns the language parameter struct associated with the data source.
			*/
			const SqlLanguage& language() const;

			/*
				execute
				Executes the specified DML statement against the data source. DML statements do not return data sets
				and include INSERT, UPDATE and DELETE statements. Returns the number of rows affected by the statement.
			*/
			virtual unsigned execute( const SqlModifyStatement& statement ) = 0;

			/*
				execute
				Executes the specified query against the data source and return the results. Queries include SELECT and
				UNION type statements. Returns the number of rows in the result set.
			*/
			virtual unsigned execute( Result& result, const SqlQueryStatement& statement, SqlQueryInfo* pinfo = NULL ) = 0;

			/*
				autoCommit
				returns whether the database will automatically commit DML statements.
			*/
			virtual bool autoCommit() = 0;

			/*
				setAutoCommit
				Enables or disables auto-commit (see above) if the data source supports transactions.
			*/
			virtual void setAutoCommit( bool bOn ) = 0;

			/*
				begin
				Starts a new transaction (if the data source supports transactions). If auto-commit is enabled, it may
				be disabled.
			*/
			virtual void begin() = 0;

			/*
				commit
				Commits the current transaction (if the data source supports transactions). NOTE if auto-commit was
				enabled prior to calling begin() and the call to begin() disabled it, this will NOT restore auto-commit
				to its previous state.
			*/
			virtual void commit() = 0;

			/*
				commit
				Rolls back the current transaction (if the data source supports transactions). NOTE if auto-commit was
				enabled prior to calling begin() and the call to begin() disabled it, this will NOT restore auto-commit
				to its previous state.
			*/
			virtual void rollback() = 0;

			/*
				schemas
				returns a list of schemata in the specified catalog. Refer to the documentation of the specific data
				source for what constitutes a schema.
			*/
			virtual void schemas( SqlSchemaList& schemaS, const SqlDatabase& catalog ) = 0;

			/*
				tables
				returns a list of tables in the specified schema. Refer to the documentation of the specific data
				source for what constitutes a schema.
			*/
			virtual void tables( SqlTableList& tableS, const SqlSchema& schema ) = 0;

			/*
				invoke
				returns a list of the columns in the specified table.
			*/
			virtual void invoke( SqlQueryInfo& info, const SqlTable& table ) = 0;

			/*
				connectStringToMap
				Converts a connect string to a ConnectionParameterMap. If nRestrict is unspecified and the connect
				string contains a plus sign (+), it will assume the delimiter is plus. Failing that, if the connect
				string contains a semi-colon (;), it will assume that the delimiter is semi-colon, otherwise it will
				assume that the delimiter is colon (:). In the event that parameters may contain either (or both) of
				these, any delimiter can be specified in nRestrict. Alternatively, the ConnectionParameterMap can be
				built manually.
				Although any of these characters may appear in a coded connect string, where the connect string is
				passed on the command line, it must abide by certain other restrictions (it cannot contain semi-
				colon, for example, as it is a separator in *NIX shells) and certain Oracle TNS free SIDs can also
				contain a colon, making it unacceptable as a separator (for example, host:port/sid).
			*/
			static void connectStringToMap( ConnectionParameterMap& map, const std::string& strConnect, int nRestrict = 0 );

		protected:
			void setLanguage( const SqlLanguage& language );
			void connect();

		private:
			DalDatabase( const DalDatabase& );
			const DalDatabase& operator=( const DalDatabase& );

			friend class DalDynamic;
	};
};

//extern "C" RFCDAL_EXPORT RfcDal::DalDatabase* dalDatabase() throw ( RfcDal::DalDatabaseException );

#endif /* !__DalDatabase_h__ */
